<?php

namespace SuperRhino\Utils\SubTable;

use SuperRhino\Utils\Cache;

trait SubTableTrait {

    public static function subTableCount(): int {
        return 32;
    }

    public static function subTableName(string $key): string {
        return static::subTableNameBySuffix(static::getSubTableSuffix($key));
    }

    public static function subTableNameBySuffix(string $suffix): string {
        return (new static)->getTable() . '_' . $suffix;
    }

    public static function isSuffixRecord() {
        return true;
    }

    public static function getSuffixRecordTableName(): string {
        return (new static)->getTable() . '_suffix_record';
    }

    public static function subTableSuffixCacheKey($key): string {
        return 'STSuffix:' . static::getSuffixRecordTableName() . ':' . strval($key);
    }

    public static function subTableSuffixCacheTimeout(): int {
        return 3 * 60 * 60;
    }

    public static function getSubTableSuffixList(): array {
        $list = [];
        for ($i = 1; $i <= static::subTableCount(); $i++) {
            $list[] = self::formatSubTableNum($i);
        }
        return $list;
    }

    public static function getSubTableSuffix($key, bool $isCache = true): string {
        if (!static::isSuffixRecord()) {
            return self::generateSubTableSuffix($key);
        }

        if ($isCache) {
            $suffix = Cache::get(static::subTableSuffixCacheKey($key));
            if ($suffix) {
                return $suffix;
            }
        }

        $suffix = self::getSubTableSuffixForDB($key);
        if ($suffix) {
            Cache::set(static::subTableSuffixCacheKey($key), $suffix, static::subTableSuffixCacheTimeout());
            return $suffix;
        }

        //生成前缀
        $suffix = self::generateSubTableSuffix($key);
        //插入数据库
        $result = self::setSubTableSuffixToDB($key, $suffix);
        if (!$result) {
            return self::getSubTableSuffix($key, $isCache);
        } else {
            Cache::set(static::subTableSuffixCacheKey($key), $suffix, static::subTableSuffixCacheTimeout());
            return $suffix;
        }
    }

    public static function suffixRecordTableUp() {
        if (!static::isSuffixRecord()) {
            return;
        }
        DBManager::schema()->create(static::getSuffixRecordTableName(), function ($table) {
            $table->bigIncrements('id');
            $table->string('key', 24);
            $table->integer('table_num');
            $table->timestamp('created_at')->useCurrent();
            $table->timestamp('updated_at')->useCurrent();
            $table->unique('key');
        });
    }

    public static function suffixRecordTableDown() {
        if (!static::isSuffixRecord()) {
            return;
        }
        DBManager::schema()->dropIfExists(static::getSuffixRecordTableName());
    }

    /**
     * 格式化后缀
     * @param int $num
     * @return string
     */
    private static function formatSubTableNum(int $num): string {
        return sprintf("%03d", $num);
    }

    /**
     * 生成表序号
     * @param $key
     * @return string
     */
    private static function generateSubTableSuffix($key): string {
        $num = intval(sprintf('%u', crc32(strval($key))));
        return self::formatSubTableNum(($num % static::subTableCount()) + 1);
    }

    /**
     * @param $key
     * @return false|string
     */
    private static function getSubTableSuffixForDB($key): string {
        $result = DBManager::table(static::getSuffixRecordTableName())
            ->where('key', strval($key))
            ->first();
        if (!$result) {
            return false;
        }
        return self::formatSubTableNum($result->table_num);
    }

    private static function setSubTableSuffixToDB($key, string $suffix): string {
        $rows = DBManager::table(static::getSuffixRecordTableName())->insertOrIgnore([
            'key' => strval($key),
            'table_num' => intval($suffix),
        ]);
        return $rows > 0;
    }
}